import pandas as pd
import ast

"""
Please upload any useful data cleaning functions you tend to use regularly
"""


def __turn_string_to_dict(d):
    """if already a dict return dict, if string return dict"""
    try:
        # If d is a string
        d = ast.literal_eval(d)
    except (ValueError):
        # In case d is already a dictionary data type
        pass
    return d


def check_json_col_names(df, col):
    """Checks if JSON based column has consistent varibles among itself across all rows

    Args:
        df (_type_): _description_
        col (_type_): _description_
    """

    df[col].apply(__turn_string_to_dict)
    initial_keys = df[col][0].keys()
    for _, row in df.iterrows():
        if initial_keys != row[col].keys():
            print(
                f"The JSON column '{col}' does NOT have consistent varibles across all entries"
            )
            return
    print(f"The JSON column '{col}' DOES have consistent varibles across all entries")


def turn_json_col_to_features(df, col_name):
    """In a pandas df expand a dictionary based column and drop the original"""
    # Converts column to dictionary if it isn't already
    df[col_name].apply(__turn_string_to_dict)

    for i, row in df.iterrows():

        d = row[col_name]
        # If nan value then continue
        if isinstance(d, float):
            continue

        for feature_name, feature_value in d.items():
            df.loc[i, col_name + "_" + feature_name] = str(feature_value)

    # Remove old JSON column
    df.drop(columns=col_name, inplace=True)



import time


class Stopwatch:
    def __init__(self):
        self._start_time = 0
        self._end_time = 0
        self._message = ""
        self._counter = 0
        self._get_new_sw_counter()

    def tick(self, message: str = "unspecified"):
        self._start_time = time.time()
        self._message = self._convert_to_single_line_str(message)
        # self._counter += 1

        msg = f">>> {self._counter} starting operation: {self._message}"
        print(msg)

    def tock(self):
        self._end_time = time.time()
        message = f"<<< {self._counter} time taken: {self._time_diff()} seconds"
        print(message)

    def _time_diff(self) -> float:
        time_diff = round(self._end_time - self._start_time, 2)

        return time_diff

    @staticmethod
    def _convert_to_single_line_str(message: str) -> str:
        return "\\n".join(message.splitlines())

    @staticmethod
    def _check_sw_counter():
        if "sw_counter" not in globals():
            global sw_counter
            sw_counter = 0

    def _get_new_sw_counter(self):
        self._check_sw_counter()

        global sw_counter
        sw_counter += 1

        self._counter = sw_counter



import pytest
from rowan_ds_tools_10.io import db_helper
import pandas as pd
import numpy as np
import sqlalchemy


postgress = db_helper.PostgressHelper("pytests")


def test_query():
    query = "select * from example_1"
    df = postgress.query(query)

    result = df.equals(
        pd.DataFrame(
            np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), columns=["a", "b", "c"]
        )
    )
    assert result


def test_upload_to_db_and_alter_db():

    # Upload to DB test
    df = pd.DataFrame({"calories": [420, 380, 390], "duration": [50, 40, 45]})
    postgress.upload_to_db(df, "temp_1", if_exists="replace")

    # Check df's match
    df_queried = postgress.query("select * from temp_1", dequote=False)
    assert df.equals(df_queried)
    "dataframes do not match"

    # test we can drop the table
    query = "drop table temp_1"
    postgress.alter_db(query)

    # Check the table is deleted
    with pytest.raises(sqlalchemy.exc.ProgrammingError):
        df_queried = postgress.query("select * from temp_1")


def test_query_column_names():

    assert (postgress.query_column_names("example_1") == ["a", "b", "c"]).any()
    "Column name query failed"


def test_check_columns_align():

    df = pd.DataFrame(
        np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), columns=["a", "b", "c"]
    )

    postgress.check_columns_align(df, "example_1")

    with pytest.raises(ValueError):
        postgress.check_columns_align(df, "example_1", dequote=False)



# lots of diff write_log functions, but maybe only one logging decorator?
import logging
import json
import functools
import inspect

from ._param_validation import validate_params, StrOptions

config = {"level": "info", "output_file_location": "system.log"}

level_lookup = {"debug": 10, "info": 20, "warning": 30, "error": 40, "critical": 50}

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(level_lookup[config["level"]])
handler = logging.FileHandler(config["output_file_location"], mode="w")
formatter = logging.Formatter("%(asctime)s - %(levelname)s -%(message)s")
handler.setFormatter(formatter)
LOGGER.addHandler(handler)


@validate_params(
    {"level": [StrOptions({"debug", "info", "warning", "error", "critical"})]}
)
def write_basic_log(level="info", error_message=None, **variables_to_log):
    """writes basic log when called

    Args:
        level (str): type of log
        error_message (str, optional): error messsage to record if one has been thrown up
    """

    # Create JSON message to log
    json_message = {"error_message": error_message}
    json_message = json_message | variables_to_log
    json_message = json.dumps(json_message)

    # Log the message
    LOGGER.log(level=level_lookup[level], msg=json_message)


@validate_params({"silent": ["boolean"], "log_varibles": ["boolean"]})
def function_logging_decorator(silent=True, log_varibles=True):
    """Multi-level decorator to log if function runs or not, also sets option to fail in silence
    wraps function around a try/except clause and has option to record all varibles passed to the function
    Args:
        silent (bool): If True then function will not throw error if it fails
        log_varibles (bool): If True then the input varibles to the function will be logged
    """

    # Inner decorator needed so that our decorator can have arguments
    def inner_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):

            if log_varibles:
                all_func_args = inspect.signature(func).bind(*args, **kwargs).arguments
                all_func_args = dict((k, str(v)) for k, v in all_func_args.items())
            else:
                all_func_args = {}

            # Try run fuction, and log as 'info' if it succeeds at running, along with pre-defined varibles to log **variables_to_log
            try:
                out = func(*args, **kwargs)
                write_basic_log(level="info", **all_func_args)
                return out

            # If function fails log as error and record error message, along with pre-defined varibles to log **variables_to_log
            except Exception as e:
                write_basic_log(level="error", error_message=repr(e), **all_func_args)
                if silent:
                    return
                else:
                    raise

        return wrapper

    return inner_decorator


# def environment_logging_decorator(self, silent=True, **variables_to_log):
#     """Multi-level decorator to log if function runs or not, also sets option to fail in silence
#     Use this decorator if you wish to record specific varibles in environment, and not just variables passed to function

#     Args:
#         silent (bool): If True then function will not throw error if it fails
#         **variables_to_log: set these as varibles in current workspace you wish to record in the logs
#     """

#     # Inner decorator needed so that our decorator can have arguments
#     def inner_decorator(func):

#         @functools.wraps(func)
#         def wrapper(*args,**kwargs):

#             # Try run fuction, and log as 'info' if it succeeds at running, along with pre-defined varibles to log **variables_to_log
#             try:
#                 out = func(*args,**kwargs)
#                 self.write_basic_log(level='info', **variables_to_log)
#                 return out

#             # If function fails log as error and record error message, along with pre-defined varibles to log **variables_to_log
#             except Exception as e:
#                 self.write_basic_log(level='error', error_message=repr(e), **variables_to_log)
#                 if silent:
#                     return
#                 else:
#                     raise

#         return wrapper

#     return inner_decorator



import pandas as pd


def isOneToOne(df, col1, col2):
    """Checks if col1 and col2 have a 1-1 relationship in a pandas df"""
    first = df.drop_duplicates([col1, col2]).groupby(col1)[col2].count().max()
    second = df.drop_duplicates([col1, col2]).groupby(col2)[col1].count().max()
    return first + second == 2



import functools
import time

import pandas as pd
import sqlalchemy as sqa

from ..utils._param_validation import validate_params


def _db_connector_decorator(func):
    """Decorator which

    1. sets up connections to the db
    2. peforms decorated function
    3. closes connections to the DB

    Args:
        func (function): function to interact with the DB

    Returns:
        _type_: _description_
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):

        # ---------- Sets up DB connection ----------
        try:
            engine = sqa.create_engine(self.connstring)
            conn = engine.connect()
        except:
            # Sometimes it doesn't connect on first try
            time.sleep(4)
            engine = sqa.create_engine(self.connstring)
            conn = engine.connect()

        # ---------- DB interaction ----------
        try:
            out = func(self, *args, **kwargs, conn=conn)

        except Exception as e:
            raise (e)

        # ---------- Terminate db connection ----------
        finally:
            conn.close()
            engine.dispose()

        return out

    return wrapper


def strip_column_wrappers(df):
    return [x.replace('"', "") for x in df.columns]


def add_column_wrappers(df):
    return ['"' + x + '"' for x in df.columns]


class PostgressHelper:
    @validate_params({"conn_string": [str]})
    def __init__(self, conn_string):
        self.connstring = conn_string

    def get_db_engine(self):

        engine = sqa.create_engine(self.connstring)
        return engine

    @_db_connector_decorator
    @validate_params({"dequote": ["boolean"], "query": [str]})
    def query(self, query, dequote=True, **kwargs):
        """Queries the postgress DB and return the quieried result

        Args:
            query (str): query string
            dequote (bool, optional): option to dequote columns or not. Defaults to True.

        Returns:
            _type_: _description_
        """
        df = pd.read_sql(query, con=kwargs["conn"])
        if dequote:
            df.columns = strip_column_wrappers(df)

        return df

    @_db_connector_decorator
    @validate_params({"query": [str]})
    def alter_db(self, query, **kwargs):
        """Peforms query to alter the postgress DB, such as drop drop table or rows

        Args:
            query (str): query to alter db
        """
        kwargs["conn"].execute(query)
        print("query executed")

        return

    @_db_connector_decorator
    @validate_params({"df": [pd.core.frame.DataFrame], "table_name": [str]})
    def upload_to_db(
        self,
        df,
        table_name,
        if_exists,
        index=False,
        index_label=None,
        chunksize=None,
        dtype=None,
        method=None,
        add_quotes=True,
        **kwargs,
    ):

        """Uploads data to the database

        Args:
            df (pd.core.frame.DataFrame): dataframe to upload
            table_name (str): name of table to upload into
            if_exists (str): .to_sql arg
            index (bool): .to_sql arg
            index_label (str): .to_sql arg
            chunksize (int): .to_sql arg
            dtype (): .to_sql arg
            method (): .to_sql arg
            add_quotes (bool): whether to add quotes to df in ordder to insert into pandas df


        Raises:
            ValueError: _description_

        Returns:
            _type_: _description_
        """
        # TODO add if statements on if_exists to add this in
        # self.check_columns_align(df, table_name, dequote=add_quotes)
        if add_quotes:
            df.columns = add_column_wrappers(df)

        df.to_sql(
            table_name,
            con=kwargs["conn"],
            if_exists=if_exists,
            index=index,
            index_label=index_label,
            chunksize=chunksize,
            dtype=dtype,
            method=method,
        )
        print("upload completed")

        return

    @validate_params({"table_name": [str]})
    def query_column_names(self, table_name, dequote=True):
        """queries postgress DB to get the column names from a specified table

        Args:
            table_name (str): name of table to query

        Returns:
            (list): list of column names in table
        """

        column_name_query = (
            "SELECT column_name FROM information_schema.columns where table_name = '"
            + table_name
            + "' order by column_name"
        )
        if dequote:
            column_names = (
                self.query(column_name_query)["column_name"]
                .apply(lambda x: x.replace('"', ""))
                .values
            )
        else:
            column_names = self.query(column_name_query)["column_name"].values

        return column_names

    @validate_params({"df": [pd.core.frame.DataFrame], "table_name": [str]})
    def check_columns_align(self, df, table_name, dequote=True):
        """Checks names in local df matches names in postgres table

        Args:
            df (pd.core.frame.DataFrame): local df
            table_name (str): name of table in pandas df
        """

        db_cols = set(self.query_column_names(table_name, dequote=dequote))
        df_cols = set(df.columns)

        db_cols_not_in_df = db_cols.difference(df_cols)
        df_cols_not_in_db = df_cols.difference(db_cols)

        if len(db_cols_not_in_df) + len(df_cols_not_in_db) != 0:
            raise ValueError(
                f"The following columns {db_cols_not_in_df} are in the db but not in the df \n The following columns {df_cols_not_in_db} are in the df but not in the db"
            )
        else:
            print("All columns align!")
            return
